@@ -1,5 +1,6 @@
 import math
 import numpy as np
+import copy
 from abc import ABC, abstractmethod
 
 from elsp_env_manager.base import Distribution, MID, Substance, Material, NormalDistribution
@@ -181,9 +182,12 @@ class ChangeoverSetup(Setup):
         self.changeover_cost_v = 0
 
     def needs_changeover(self, prod, t):
-        return (not prod.same_series(self.last_prod)) or (
-                self.process_unit.process_start_time - self.process_unit.last_process_start_time) >= self.get_distribution(
+        needs_changeover = (not prod.same_series(self.last_prod)) or (
+                self.process_unit.process_start_time - self.process_unit.last_process_end_time) >= self.get_distribution(
             "setup_expiry", prod).sample(t)
+        if not needs_changeover:
+            print('no need changeover')
+        return needs_changeover
 
     def _justify(self, prod: Substance, t):
         time = self.get_distribution("changeover_time", prod).sample(
@@ -191,13 +195,17 @@ class ChangeoverSetup(Setup):
                                                                             prod).sample(t)
         finish = t >= time
         if finish:
+            self.refresh_increment_inf("setup time", t)
             if self.needs_changeover(prod, t):
                 changeover_cost = self.get_distribution("changeover_cost", prod).sample(t)
                 self.cost = changeover_cost
                 self.changeover_cost_v = changeover_cost
                 self.refresh_increment_inf("changeover cost", changeover_cost)
+                self.refresh_increment_inf("changeover time", t)
             else:
                 self.cost = self.get_distribution("turnaround_cost", prod).sample(t)
+                self.refresh_increment_inf("turnaround cost", changeover_cost)
+                self.refresh_increment_inf("turnaround time", t)
         self.refresh_inf()
         return finish
 
@@ -240,13 +248,25 @@ class Launch(ProductionProcess):
         self.launch_time = launch_time
         self.launch_cost = launch_cost
         # self.obs_inf_k = "launch_" + str(self.process_unit.m_id)
+    
+    def reset(self, faild, *args):
+        if faild:
+            if len(args) > 0:
+                lead_process_finish_time = args[0]
+                self.refresh_increment_inf("launch time", self.process_unit.get_time() - lead_process_finish_time)
+            else:
+                self.refresh_increment_inf("launch time", self.process_unit.get_time())
+        super().reset()
 
     def _justify(self, prod: Substance, t):
         time = self.launch_time.sample(t)
         finish = t >= time
         if finish:
             self.cost = self.get_distribution("launch_cost", prod).sample(t)
-        self.refresh_inf()
+            self.refresh_increment_inf("launch time", t)
+            self.refresh_inf()
+        else:
+            self.refresh_inf()
         return finish
 
     def refresh_inf(self):
@@ -292,7 +312,13 @@ class Conversion(ProductionProcess):
         self.conversion_daily_cost = 0
         self.obs_inf = {}
 
-    def reset(self):
+    def reset(self, faild, *args):
+        if faild:
+            if len(args) > 0:
+                lead_process_finish_time = args[0]
+                self.refresh_increment_inf("conversion time", self.process_unit.get_time() - lead_process_finish_time)
+            else:
+                self.refresh_increment_inf("conversion time", self.process_unit.get_time())
         super().reset()
         self.conversion_prepare_cost_v = 0
         self.conversion_finish_cost_v = 0
@@ -324,6 +350,7 @@ class Conversion(ProductionProcess):
             self.cost += finish_cost
             self.conversion_finish_cost_v += finish_cost
             self.refresh_increment_inf("finish cost", finish_cost)
+            self.refresh_increment_inf("conversion time", t)
             if self.prod is None:
                 output_rate += self.get_distribution("conversion_rate", prod).sample(t)
             else:
@@ -538,9 +565,13 @@ class Sale(Process):
             total_demand_fit += d["fit"]
             d["not_fit"] -= ret
             sub = Substance.substance[MID(SUBSTANCE_ID_PREFIX + "_{}_2".format(prod))]
-            self.sales_volume += ret * self.get_distribution("price_per_unit", sub).sample(self.process_unit.get_time())
-            self.backlog_penalty += d["not_fit"] * self.get_distribution("backlog_penalty_cost_per_unit_per_day",
+            sales_volume = ret * self.get_distribution("price_per_unit", sub).sample(self.process_unit.get_time())
+            self.sales_volume += sales_volume
+            d["sales_volume"] += sales_volume
+            backlog_penalty = d["not_fit"] * self.get_distribution("backlog_penalty_cost_per_unit_per_day",
                                                                          sub).sample(self.process_unit.get_time())
+            self.backlog_penalty += backlog_penalty
+            d["backlog_penalty"] += backlog_penalty
             decay = d["not_fit"] * self.backlog_decay_per_half_year
             self.backlog_decay += decay
             d["decay"] += decay
@@ -558,6 +589,9 @@ class Sale(Process):
         decay = {}
         csl = {}
         inventory = {}
+        sales_volume = {}
+        backlog_penalty = {}
+        unit_price = {}
         for p, d in demand.demand_dict.items():
             _all[p] = d['all']
             fit[p] = d['fit']
@@ -565,13 +599,22 @@ class Sale(Process):
             decay[p] = d['decay']
             csl[p] = d['fit'] / d['all']
             inventory[p] = store.prod_total[p] if p in store.prod_total else 0
-        self.set_obs_inf(self.obs_inf_k, {"sales volume": self.sales_volume, "backlog decay": self.backlog_decay,
-                                          "backlog penalty": self.backlog_penalty,
-                                          "csl": total_demand_fit / total_demand * 100,
-                                          "all order values": _all, "fit order values": fit,
-                                          "not fit order values": not_fit,"decay values": decay,
-                                          "inventory values": inventory,
-                                          "csl s": csl, "current time":self.process_unit.get_time()})
+            sales_volume[p] = d['sales_volume']
+            backlog_penalty[p] = d['backlog_penalty']
+            unit_price[p] = d["sales_volume"] / d['fit'] if d['fit'] > 0 else 0
+        new_csl = total_demand_fit / total_demand * 100
+        last_csl = self.process_unit.env.variables.get_variable("CSL")
+        self.set_obs_inf(self.obs_inf_k,
+                         {"sales volume": self.sales_volume, "backlog decay": self.backlog_decay,
+                          "backlog penalty": self.backlog_penalty,
+                          "csl": new_csl,
+                          "delta csl": new_csl - last_csl,
+                          "all order values": _all, "fit order values": fit,
+                          "not fit order values": not_fit,"decay values": decay,
+                          "inventory values": inventory,
+                          "unit price s": unit_price,
+                          "sales volume s": sales_volume,
+                          "csl s": csl, "current time":self.process_unit.get_time()})
 
     @staticmethod
     def load(json, process_unit):
@@ -589,11 +632,14 @@ class Sale(Process):
 
 class Demand(Process):
 
+    __initial_demand_dict__ = {"all": 0, "fit": 0, "not_fit": 0,
+                               "decay": 0, "sales_volume": 0, "backlog_penalty": 0}
+
     def reset(self):
         super().reset()
         self.demand_dict = {}  # 用于保存累计值
         for prod, dist in self.demand_json.items():
-            self.demand_dict[prod] = {"all": 0, "fit": 0, "not_fit": 0, "decay": 0}
+            self.demand_dict[prod] = copy.deepcopy(self.__initial_demand_dict__)
 
     def refresh_inf(self):
         pass
@@ -616,7 +662,7 @@ class Demand(Process):
         self.demand_dict = {}  # 用于保存累计值
         for prod, dist in demand_json.items():
             self.demand_dist[prod] = Distribution.load(demand_json, prod)
-            self.demand_dict[prod] = {"all": 0, "fit": 0, "not_fit": 0, "decay": 0}
+            self.demand_dict[prod] = copy.deepcopy(self.__initial_demand_dict__)
         pass
 
     def generate(self):
@@ -635,6 +681,7 @@ class Scheduling(Process):
     def reset(self):
         super().reset()
         self.obs = []
+        self.idle_cnt = 0
         for i in range(self.prod_feature_dim * len(self.scheduling_valid_prod) + self.env_feature_dim):
             self.obs.append(0)
         return self.obs
@@ -660,6 +707,7 @@ class Scheduling(Process):
         self.last_action = None
         self.action = None
         self.obs = []
+        self.idle_cnt = 0
         for i in range(self.prod_feature_dim * len(self.scheduling_valid_prod) + self.env_feature_dim):
             self.obs.append(0)
 
@@ -669,7 +717,7 @@ class Scheduling(Process):
         self.delta_time = 0
         self.delta_profit = 0
         self.obs = []
-        self.last_action = action
+        self.last_action = 0 if self.action is None else self.action
         self.action = action
 
     def on_scheduling_finished(self):
@@ -680,7 +728,7 @@ class Scheduling(Process):
         action = self.action
         delta_time = self.delta_time
         self.set_obs_inf(self.obs_inf_k, {"delta profit": self.delta_profit, "last_action":last_action,"action":action,
-                                          "delta time":delta_time})
+                                          "delta time":delta_time}, replace=False)
         # self.set_obs_inf(self.obs_inf_k, {"all order values": _all, "fit order values": fit,
         #                                   "not fit order values": not_fit,"decay values": decay,
         #                                   "inventory values": inventory,"delta profit": self.delta_profit,
@@ -702,11 +750,16 @@ class Scheduling(Process):
             self.refresh_obs_inf_keys()
             self.process_unit.get_latter_unit().process(action)
             self.process_unit.env.time += 1
+            if action == 0:
+                self.idle_cnt += 1
             self.set_obs_inf("scheduling", {"time": self.process_unit.env.time})
+            self.set_obs_inf(self.obs_inf_k, {'idle percent': self.idle_cnt / self.process_unit.env.time}, replace=False)
             done = self.process_unit.get_time() >= self.get_distribution("scheduling_period", None).sample()
             if self.process_unit.needs_scheduling() or done:
                 self.on_scheduling_finished()
-                return self.obs, self.process_unit.env.variables.get_variable(self.reward_define), done
+                self.set_obs_inf(self.obs_inf_k, {"p": action}, replace=False)
+                reward = self.process_unit.env.variables.get_variable(self.reward_define)
+                return self.obs, reward, done
             return self.obs, 0, done
         else:
             raise NotImplementedError
